require 'json'

module UU
  module OS
    module REST

      # This data structure should be used for paging large result sets. It is
      # typically used as return value for uuAPI services. The corresponding DTO
      # structure with criteria should be used as parameter for the service.
      class ResultList < Array

        # Total size of the list (sum of all pages).
        # @param [Numeric]
        # @return [Numeric]
        attr_accessor :total_size

        # Criteria used to get list.
        # @param [UU::OS::REST::PagingCriteria]
        # @return [UU::OS::REST::PagingCriteria]
        attr_accessor :criteria

        # Creates new instance of result list.
        #
        # @param [Class] criteria_cls  Criteria DTO class
        # @param [Class] entry_cls Entry DTO class
        # @param [String, Hash] data JSON or +Hash+ containing result list data
        def initialize(criteria_cls, entry_cls, data=nil)
          super(0)
          if (criteria_cls.nil?) || (entry_cls.nil?)
            raise ArgumentError.new('Criteria and Entry classes must be given')
          end
          if data.nil?
            return
          elsif data.kind_of?String
            data = JSON.parse(data, :symbolize_names => true)
          end
          if !data.kind_of?Hash
            raise ArgumentError.new("Parameter data must be JSON String or Hash, but was #{data.class}.")
          end
          if data[:total_size]
            @total_size = data[:total_size]
          else
            @total_size = data[:totalSize]
          end
          @criteria = criteria_cls.new(data[:criteria])
          if data[:page_entries]
            data[:page_entries].each do |item|
              self << entry_cls.new(item)
            end
          elsif data[:pageEntries]
            data[:pageEntries].each do |item|
              self << entry_cls.new(item)
            end
          end
        end

        # @private
        def to_s
          str = "{ \"#{self.class.name}\" : { "
          str << "\"total_size\" : \"#{self.total_size}\" , "
          str << "\"criteria\" : #{self.criteria.to_s} , "
          str << "\"page_entries\" : [ "
          if self.size <= 3
            self.each do |item|
              if (item.kind_of?UU::OS::REST::DTO) && (!item.kind_of?UU::OS::UESURI)
                str << "#{item.to_s} , "
              else
                str << "\"#{item.to_s}\" , "
              end
            end
            str = str[0..-4] if self.size > 0
          else
            for i in 0..2
              item = self[i]
              if (item.kind_of?UU::OS::REST::DTO) && (!item.kind_of?UU::OS::UESURI)
                str << "#{item.to_s} , "
              else
                str << "\"#{item.to_s}\" , "
              end
            end
            str << "\"... and #{self.size - 3} more.\""
          end
          str << ' ] } }'
          return str
        end

        # @private
        def inspect
          # Override default inspect to suppress printing of
          # all items of list.
          "#<#{self.class.name}:#{self.object_id}>"
        end

      end

    end
  end
end
