require 'base64'
require 'uu/os/env/environment'
require 'uu/os/qos/qos_handler'
require 'uu/os/rest/remote_client'
require 'uu/os/uesuri'
require 'uu/os/security/encryption'
require 'uu/os/security/session_app_identity_attributes'

module UU
  module OS
    module Security

      # Service representing session.
      class Session
        extend UU::OS::Security::Encryption

        # One of possible values for the :context_scope parameter, see {UU::OS::Security::Session.login login} method.
        # If this value is used (for the :context_scope parameter), an identity is set to authentication context of the current
        # thread after successful {UU::OS::Security::Session.login login} invocation. This value is the default for
        # the +context_scope+ parameter if no value is specified.
        #
        # @example
        #   UU::OS::Security::Session.login(
        #     user_credentials: ['access_code_1', 'access_code_2'],
        #     context_scope: UU::OS::Security::Session::CONTEXT_SCOPE_THREAD
        #   )
        CONTEXT_SCOPE_THREAD = :context_scope_thread;

        # One of possible values for the :context_scope parameter, see {UU::OS::Security::Session.login login} method.
        # If this value is used (for the :context_scope parameter), the authentication context of the current thread will
        # not be influenced after successful {Session#login login} method invocation. The {Session#login login} method only
        # returns a session instance which may be used for authentication of other calls (e.g. passing the session as
        # parameter for {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient} instance creation).
        #
        # @example
        #   UU::OS::Security::Session.login(
        #     user_credentials: ['access_code_1', 'access_code_2'],
        #     context_scope: UU::OS::Security::Session::CONTEXT_SCOPE_LOCAL
        #   )
        CONTEXT_SCOPE_LOCAL = :context_scope_local;

        # Service path
        # @private
        PATH = "ues/core/security/session/UESSession"

        # Handshake service path
        # @private
        @@HANDSHAKE_PATH = 'ues/platform/security/authn/UESHandshakeService'

        # Name of thread parameter for storing authentication tokens
        # @private
        @@AUTHN_TOKEN_PARAM = :'uu.os.rest.remote-client-authn-token'

        # Value returned by REST API in case of null result
        # @private
        @@NULL_URI_VALUE = '{"nil":"true"}'

        # Default trusted realm
        # @private
        DEFAULT_TRUSTED_REALM = 'trusted.uuee.uu'

        # Default untrusted realm
        # @private
        DEFAULT_UNTRUSTED_REALM = 'uuee.uu'

        # Name of configuration property, whose value overwrite default trusted realm
        # @private
        TRUSTED_REALM_CONF_PAR_NAME = 'uu.os.security.appsecret.trusted-realm'

        # Name of configuration property, whose value overwrite default trusted realm
        # @private
        UNTRUSTED_REALM_CONF_PAR_NAME = 'uu.os.security.appsecret.untrusted-realm'

        # Name of configuration property for definition of password files directory
        # @private
        PASSWORD_FILES_DIR_PAR_NAME = 'uu.os.security.password-files-dir'

        # Relative path to file in war, where app secret is stored
        # @private
        SECRET_FILE_PATH = 'META-INF/uu.os.appSecret'

        # Lower-cased JWT prefix.
        # @private
        @@JWT_TOKEN_PREFIX = 'bearer ';

        # JWT token parts separator.
        # @private
        @@JWT_TOKEN_PARTS_SEPARATOR = '.';

        # Hash attribute name which holds application credentials.
        # @private
        @@APP_CREDENTIALS = :app_credentials;

        # Hash attribute name which holds user credentials.
        # @private
        @@USER_CREDENTIALS = :user_credentials;

        # Name of Hash parameter which holds current credentials. This hash is
        # used for transferring parameters during POST REST request.
        # @private
        @@CURRENT_CREDENTIALS = :current_credentials;

        # Claim name which contains client Universe ID.
        # @private
        @@CLIENT_UID = 'client_uid';

        # Claim name which contains client application code.
        # @private
        @@CLIENT_CODE = 'client_code';

        # Log user in.
        #
        # @param [*String, Hash] credentials Valid parameters are:
        #   * Security realm, access code 1, access code 2
        #   * access code 1, access code 2 (uses default security realm)
        #   * Path to password file (absolute or relative to uu home)
        #   * HTTP authorization token
        #   * Hash with user (:user_credentials) and application (:app_credentials)
        #     credentials. Both values are optional and if specified, it can be
        #     formed from parameters valid for non-hash parameter variants.
        #     If no application credentials are given, the system will check whether
        #     exist the <%uu-home%>/app_identity file and loads application
        #     credentials from the file.
        #
        #     The next optional hash parameter is context scope (:context_scope). This parameter affects settings of
        #     the authentication context after successful login method call.
        #     If the context scope parameter is set to the {CONTEXT_SCOPE_THREAD CONTEXT_SCOPE_THREAD} value, an identity will be set on
        #     the authentication context of the current thread.
        #     If the value is set to the {CONTEXT_SCOPE_LOCAL CONTEXT_SCOPE_LOCAL}, a session instance will be returned. The session instance
        #     may be used for authentication of next calls (e.g. passing the session as parameter for
        #     {UU::OS::CMD::CommandClient UU::OS::CMD::CommandClient} instance creation).
        #
        # @return [TrueClass,UU::OS::Security::Session] +true+ value if the :context_scope parameter is set to the {CONTEXT_SCOPE_THREAD CONTEXT_SCOPE_THREAD} value;
        #     {UU::OS::Security::Session Session} instance if the :context_scope parameter is set to the {CONTEXT_SCOPE_LOCAL}
        #
        # @example
        #   credentials = Hash.new
        #   credentials[:user_credentials] = ['accessCode1', 'accessCode2']
        #   credentials[:app_credentials] = ['/etc/password_file']
        #   credentials[:context_scope] = UU::OS::Security::Session::CONTEXT_SCOPE_LOCAL
        #   UU::OS::Security::Session.login(credentials)
        def self.login(*credentials)
          is_thread_scope = CONTEXT_SCOPE_THREAD == get_context_scope(credentials);

          session = process_login(get_authn_token(), *credentials);
          session.send(:previous_session, load_session());
          if (is_thread_scope)
            store_session(session);
          end

          return is_thread_scope ? true : session;
        end

        # Determines context scope from parameters passed to the {Session#login login} method.
        # @param [Array] array of parameters passed to the {Session#login login} method
        # @return [Symbol] return the specified context scope or {CONTEXT_SCOPE_THREAD CONTEXT_SCOPE_THREAD} as the default value if not specified
        # @private
        def self.get_context_scope(parameters)
          result = nil;
          if (parameters.kind_of?(Array) && parameters[0].kind_of?(Hash))
            param = parameters[0][:context_scope];
            if (param != nil && param != CONTEXT_SCOPE_THREAD && param != CONTEXT_SCOPE_LOCAL)
              raise "Unexpected value of :context_scope parameter (#{param}). The expected value is either UU::OS::Security::Session::CONTEXT_SCOPE_LOCAL or UU::OS::Security::Session::CONTEXT_SCOPE_THREAD or nil (i.e. not specified).";
            end
            result = param;
          end
          if (result == nil)
            result = CONTEXT_SCOPE_THREAD; # the default value is CONTEXT_SCOPE_THREAD
          end
          return result;
        end
        private_class_method :get_context_scope;

        # Prepares login parameters and invokes appropriate login variant.
        # @param [String, nil] authn_token current authentication token
        # @param [Hash, *String] credentials credentials used during login
        # @private
        def self.process_login(authn_token, *credentials)
          newToken = nil;

          app_crd = get_app_credentials(*credentials)
          if ( app_crd != nil )
            if ( credentials[0].kind_of?(Hash))
              credentials[0][@@APP_CREDENTIALS] = app_crd;
              newToken = app_login(authn_token, *credentials);
            else
              hash = Hash.new
              hash[@@USER_CREDENTIALS] = credentials;
              hash[@@APP_CREDENTIALS] = app_crd;
              newToken = app_login(authn_token, hash);
            end
          else
            usr_crd = credentials;
            if usr_crd[0].kind_of?(Hash)
              usr_crd = usr_crd[0][@@USER_CREDENTIALS];
            end
            newToken = user_login(authn_token, *usr_crd);
          end

          session = self.send(:new, newToken);
          return session;
        end
        private_class_method :process_login;

        # Log in user (without application).
        # @param [String, nil] authn_token current authentication token
        # @param [Hash, *String] credentials credentials used during login
        # @private
        def self.user_login(authn_token, *credentials)
          tmpToken = get_access_token(*credentials)

          # Evaluate token, and obtain new JWT token
          # Send also previous token (if available) to create auth chain
          svc = UU::OS::REST::RemoteClient.new(Session)
          svc.auth_token=tmpToken
          result = svc.post(:login, nil, authn_token).gsub('"', '')

          # Obtaining of JWT token passed - We have valid token
          return result
        end
        private_class_method :user_login;

        # Log in user and application.
        # @param [String, nil] authn_token current authentication token
        # @param [Hash, *String] credentials credentials used during login
        # @private
        def self.app_login(authn_token, credentials)
          tmp = *credentials[@@USER_CREDENTIALS];
          user_token = credentials[@@USER_CREDENTIALS] != nil ? get_access_token(*credentials[@@USER_CREDENTIALS]) : authn_token;
          if ( user_token == nil )
            raise "Login failed. There is neither authenticated user nor user credentials specified.";
          end
          app_token = credentials[@@APP_CREDENTIALS] ? get_access_token(*credentials[@@APP_CREDENTIALS]) : nil;

          payload = Hash.new
          payload[@@CURRENT_CREDENTIALS] = authn_token; # it can be null
          payload[@@APP_CREDENTIALS] = app_token

          svc = UU::OS::REST::RemoteClient.new(Session)
          svc.auth_token=user_token
          result = svc.post(:login, nil, payload.to_json()).gsub('"', '')

          return result;
        end
        private_class_method :app_login;

        # Determines application credentials.
        # @param [Hash, *String] credentials credentials used during login
        # @private
        def self.get_app_credentials(*credentials)
          result = credentials[0].kind_of?(Hash) ? result = credentials[0][@@APP_CREDENTIALS] : nil;
          if ( result == nil )
            filepath = File.join(password_files_dir, 'app_identity') # based on convention in UNI-BT:44191586422519564
            if ( File.file?(filepath) )
              result = *filepath;
            end
          end
          return result;
        end
        private_class_method :get_app_credentials

        # Verify the specified authentication token via calling handshake
        # web-service. As opposite to the login() method, this method does not
        # change authentication token.
        #
        # @param [String] token authentication token to be verified via web-service call
        # @return [true, false] true if the token is valid; false otherwise
        def self.verify(token)
          result = true;
          begin
            # TODO Until UDS and AppLog relies on application URI in JWT
            # token, we cannot generate new chained JTW token replacing
            # existing one. So we only evaluate original token.
            svc = UU::OS::REST::RemoteClient.new(UU::OS::Security, @@HANDSHAKE_PATH)
            svc.auth_token=token
            svc.get(:handshake, nil)
          rescue => e
            result = false;
          end
          return result;
        end

        # Performs login and initiates a session.
        # @param token [String] access token used for authentication (e.g. in middleware)
        # @private
        def self.init_session(token)
          login(token);
          # the token is verified now
          session = UU::OS::Security::Session.current_session();
          session.send(:set_incoming_authn_token, token);
        end

        # @private
        def self.is_jwt(token)
          return token && token.to_s.downcase.start_with?(@@JWT_TOKEN_PREFIX);
        end
        private_class_method :is_jwt;

        # Check if user is logged in.
        def self.logged_in?
          if (Thread.current[@@AUTHN_TOKEN_PARAM]) && (Thread.current[@@AUTHN_TOKEN_PARAM].size > 0)
            true
          else
            false
          end
        end

        # Returns personal role URI of the authenticated user.
        #
        # @return [UU::OS::UESURI] URI of personal role
        def self.get_personal_role
          svc = UU::OS::REST::RemoteClient.new(Session)
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.get(:getPersonalRole)
            return UU::OS::UESURI.new(res)
          end
        end

        # Returns personal role URI of user who initiated session (did the first login).
        #
        # @return [UU::OS::UESURI] URI of initiator personal role
        def self.get_initiator_personal_role
          svc = UU::OS::REST::RemoteClient.new(Session)
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.get(:getInitiatorPersonalRole)
            return UU::OS::UESURI.new(res)
          end
        end

        # Returns access role URI of authenticated user in given territory.
        # May return +nil+ in case user has no access role in territory.
        #
        # @param [String, UU::OS::UESURI] territory_uri URI of territory for which to get access role
        # @return [UU::OS::UESURI] URI of access role or +nil+
        def self.get_access_role(territory_uri)
          svc = UU::OS::REST::RemoteClient.new(Session)
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.get(:getAccessRole, territory_uri)
            if (res && res != @@NULL_URI_VALUE)
              return UU::OS::UESURI.new(res)
            else
              return nil
            end
          end
        end

        # Returns access role URI of user who initiated session in given territory.
        # May return +nil+ in case user has no access role in territory.
        #
        # @param [String, UU::OS::UESURI] territory_uri URI of territory for which to get access role
        # @return [UU::OS::UESURI] URI of initiator access role or +nil+
        def self.get_initiator_access_role(territory_uri)
          svc = UU::OS::REST::RemoteClient.new(Session)
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.get(:getInitiatorAccessRole, territory_uri)
            if (res && res != @@NULL_URI_VALUE)
              return UU::OS::UESURI.new(res)
            else
              return nil
            end
          end
        end

        # Log user out.
        #
        # @param [TrueClass, FalseClass] all If true, logout removes all tokens (otherwise only the last one)
        def self.logout(all = false)
          # TODO We should also call logout on server once service is implemented.
          if all
            Thread.current[@@AUTHN_TOKEN_PARAM] = nil
          elsif self.logged_in?
            Thread.current[@@AUTHN_TOKEN_PARAM][-1] = nil
            Thread.current[@@AUTHN_TOKEN_PARAM].compact!
          end
          return
        end

        # This method is obsoleted but it is invoked outside of this class from
        # some applications (e.g. object store, app log).
        # @private
        def self.put_authn_token(token)
          if ( token != nil )
            session = self.send(:new, token);
            store_session(session);
          end
        end
        private_class_method :put_authn_token

        # @private
        def self.store_session(session)
          if !Thread.current[@@AUTHN_TOKEN_PARAM]
            Thread.current[@@AUTHN_TOKEN_PARAM] = []
          end
          Thread.current[@@AUTHN_TOKEN_PARAM] << session
        end
        private_class_method :store_session

        # @private
        def self.load_session()
          result = nil;
          if self.logged_in?
            # Returns last verified authentication token.
            result = Thread.current[@@AUTHN_TOKEN_PARAM][-1]
          end
          return result;
        end
        private_class_method :load_session;

        # @private
        def self.get_authn_token
          session = load_session();
          return session != nil ? session.access_token : nil;
        end
        private_class_method :get_authn_token;

        # Methods build access token from given credentials.
        # @private
        def self.get_access_token(*credentials) # :nodoc:
          # TODO Since there are many problems with masking of user input, interactive login prompt is temporarily disabled
          #if (credentials.nil?) || (credentials.size == 0)
          #  # Read credentials from user input
          #  begin
          #    Timeout::timeout(120) do
          #      credentials = get_credentials
          #    end
          #  rescue Timeout::Error
          #    raise 'Access codes were not provided in given time of 2 minutes.'
          #  end
          #end
          # Check input parameters
          if (credentials.nil?) || (credentials.size < 1) || (credentials.size > 3)
            raise ArgumentError.new('Wrong number of arguments. Valid parameters are "securityRealm, accessCode1, accessCode2", "accessCode1, accessCode2", "authToken" or "passwdFile"')
          end
          if credentials.size == 1  # Input is auth token or password file
            path = credentials[0]
            if !File.exists?path
              path = File.join(password_files_dir, path)
            end
            if File.file?path # password file
              crd = UU::OS::Env::Environment.load_configuration(path)
              unless crd['accessCode1']
                # the file probably contains encrypted data
                crd = decrypt_passwd_file(path)
              end
              acc1 = crd['accessCode1']
              acc2 = crd['accessCode2']
              realm = crd['securityRealm']

              if realm.nil?
                tmpToken = acc1 << ':' << acc2
              else
                acc1 = acc1.gsub('@', '\@')
                # if realm is "untrusted" => try to load app secret and switch to trusted
                app_secret = get_app_secret if realm == app_secret_untrusted_realm
                if app_secret.nil?
                  tmpToken = acc1 << '@' << realm << ':' << acc2
                else
                  tmpToken = acc1 << '@' << app_secret_trusted_realm << ':' << acc2 << app_secret
                end
              end
              tmpToken = "Basic #{Base64.strict_encode64(tmpToken)}"
            else # auth token
              tmpToken = credentials[0]
            end
          elsif credentials.size == 2 # Input is access code 1 and 2
            acc1 = credentials[0].to_s.dup.gsub('@', '\@')
            acc2 = credentials[1].to_s.dup
            tmpToken = "Basic #{Base64.strict_encode64(acc1 << ':' << acc2)}"
          elsif credentials.size == 3 # Input is security realm, access code 1 and 2
            realm = credentials[0].to_s.dup
            acc1 = credentials[1].to_s.gsub('@', '\@')
            acc2 = credentials[2].to_s.dup

            # if realm is "untrusted" => try to load app secret and switch to trusted
            app_secret = get_app_secret if realm == app_secret_untrusted_realm
            if app_secret.nil?
              tmpToken = acc1 << '@' << realm << ':' << acc2
            else
              tmpToken = acc1 << '@' << app_secret_trusted_realm << ':' << acc2 << app_secret
            end
            tmpToken = "Basic #{Base64.strict_encode64(tmpToken)}"
          end
          return tmpToken
        end
        private_class_method :get_access_token

        # Method reads credentials from user input.
        #
        # @return User credentials
        #
        # @private
        def self.get_credentials
          credentials = []
          if (RUBY_PLATFORM[/java/]) && (Java.const_defined?:JavaxSwing)
            # We are running JRuby with jirb_swing (or possibly some other GUI console)
            # We have to use graphical login prompt because it is not possible to simply
            # read user input with masking of access codes.
            raise 'Interactive login prompt is not supported in jirb_swing console'
          end
          # We are running Ruby/JRuby in command line
          puts '======== uuOS Login ========'
          # Security Realm not yet supported
          # puts 'Security Realm (optional):'
          # val = gets.chomp # no masking necessary, simple gets is multiplatform
          # credentials[0] = val if (!val.nil?) && (val != '')
          puts 'Access Code 1 (input hidden):'
          val = get_hidden_input
          credentials[credentials.size] = val
          puts 'Access Code 2 (input hidden):'
          val = get_hidden_input
          credentials[credentials.size] = val
          return credentials
        end
        private_class_method :get_credentials

        # Reads hidden input from console based on used platform - Jruby, Windows, Linux or OSX.
        # (there is no multiplatform solution for this, even highline library does not work)
        #
        # @return User input
        #
        # @private
        def self.get_hidden_input
          input = nil
          # Check Ruby platform (input must be handled separately for JRuby, Windows and Linux/OSX)
          if RUBY_PLATFORM[/java/]
            # On java we should use system console. But we must not use readPassword method, because
            # it does not work in interactive JRuby shell (it reads single char instead of whole line
            # if typing slowly, or part of line if typing fast - it is absolutely not reliable)
            cr = Java::JavaLang::System.console.reader
            so = Java::JavaLang::System.out
            input = []
            while c = cr.read
              # Mask input by immediate delete of output by sending backspace, blank char and another backspace.
              # (but some key presses sends multiple chars to console, so delete more than single character)
              so.print("\b\b\b\b\b     \b\b\b\b\b")
              if c == 27 # escape, but also delete on linux (delete causes "27 91 51 126" sequence)
                # escape is ignored so read next char immediately
                c = cr.read
                if c == 91 # it seams we have delete sequence
                  c = cr.read
                  if c == 51 # ..still it seems like delete sequence
                    c = cr.read
                    if c == 126 # yes it was delete sequence
                      input.delete_at(-1)
                      next
                    end
                  end
                end
              end
              break if c == 13 || c == 10 # CR or NL - input is finished
              if c == 127 || c == 8 # delete or backspace
                input.delete_at(-1)
              else
                input << c
              end
            end
            # Convert input char array to string
            input = Java::JavaLang::String.new(input.to_java(:char)).to_s
          elsif RbConfig::CONFIG['host_os'].downcase[/win/]
            # On windows we have to use native library for input (windows shell does not handle tty manipulations).
            # Based on https://github.com/treasure-data/td/commit/e8f2c705fef5454fce4a70c60235073254a3d0cd
            require 'Win32API'
            input = ''
            while c = get_char_win
              next if c == 27 # escape is ignored
              break if c == 13 || c == 10 # CR or NL - input is finished
              if c == 127 || c == 8 # delete or backspace
                input.slice!(-1, 1)
              else
                input << c.chr
              end
            end
          else
            # On Linux or OSX should work standard ruby io/console library.
            require 'io/console'
            input = STDIN.noecho {|i| i.gets}.chomp
          end
          return input
        end
        private_class_method :get_hidden_input

        # @private
        def self.get_char_win
          Win32API.new('msvcrt', '_getch', [], 'L').Call
        rescue Exception
          Win32API.new('crtdll', '_getch', [], 'L').Call
        end
        private_class_method :get_char_win

        # @private
        def self.decrypt_passwd_file(path)
          file_contents = File.binread(path) # read in binary mode to prevent dropping CRLF EOLs
          temp = decrypt(file_contents)
          crd = UU::OS::Env::Environment.parse_properties(temp)
        rescue
          msg = "Unable to read credentials from password file #{File.basename(path)}. "
          msg << "Its content is either invalid or encrypted with PASSWD_FILE_GENERATOR "
          msg << "tool which uses a different encryption mechanism. Use password file "
          msg << "encrypted with UU::OS::Security::Encryption#encrypt(data) if you want "
          msg << "to use encrypted password file with uuAPI."
          raise UU::OS::Security::AuthenticationException, msg
        end
        private_class_method :decrypt_passwd_file

        # Returns directory where password files are placed
        # @private
        def self.password_files_dir
          # try to get directory from system environment (works only in jruby)
          @password_files_dir ||= Java::JavaLang::System.getProperty(PASSWORD_FILES_DIR_PAR_NAME) if RUBY_PLATFORM[/java/]
          # try to get directory from config file
          @password_files_dir ||= UU::OS::Env::Environment.load_configuration(UU::OS::Env::Environment::CFG_FILE_PATH)[PASSWORD_FILES_DIR_PAR_NAME]
          # use uuHome as default
          @password_files_dir ||= UU::OS::Env::Environment.uu_home
        end
        private_class_method :password_files_dir

        # Returns trusted realm
        # @private
        def self.app_secret_trusted_realm
          # try to get trusted realm from config file
          @trusted_realm ||= UU::OS::Env::Environment.load_configuration(UU::OS::Env::Environment::CFG_FILE_PATH)[TRUSTED_REALM_CONF_PAR_NAME]
          # get default trusted realm
          @trusted_realm ||= DEFAULT_TRUSTED_REALM
        end
        private_class_method :app_secret_trusted_realm

        # Returns untrusted realm
        # @private
        def self.app_secret_untrusted_realm
          # try to get trusted realm from config file
          @untrusted_realm ||= UU::OS::Env::Environment.load_configuration(UU::OS::Env::Environment::CFG_FILE_PATH)[UNTRUSTED_REALM_CONF_PAR_NAME]
          # get default trusted realm
          @untrusted_realm ||= DEFAULT_UNTRUSTED_REALM
        end
        private_class_method :app_secret_untrusted_realm

        # Search for the file in war on path specified by the {UU::OS::Security::Session::SECRET_FILE_PATH} constant and returns it's content.
        # @private
        def self.get_app_secret
          # if path does not contain WEB-INF => we are not in .war => return nil
          return nil unless  __FILE__.include? 'WEB-INF'

          file_path = File.join('..', SECRET_FILE_PATH)

          if  File.exists? file_path
            # pwd is in WEB-INF and secret file exists

            file = File.open(file_path, 'rb')
          else
            # pwd is somewhere else or secret file do not exists

            # get war root
            war_root = __FILE__.split('WEB-INF')[0]

            # concat file path
            file_path = war_root + SECRET_FILE_PATH

            # secret file does not exists
            return nil unless File.exists? file_path

            file = File.open(file_path, 'rb')
          end
          return file.read
        end
        private_class_method :get_app_secret

        # Returns identity of the current application. Returns nil if the application identity is not available.
        # @return [UU::OS::Security::SessionAppIdentityAttributes] Object containing attributes of the server application identity or nil if the application identity is not available.
        def self.get_app_identity_attributes
          return get_app_identity_attrs(get_authn_token);
        end

        # Returns identity of the client application. Returns nil if the application identity is not available.
        # @return [UU::OS::Security::SessionAppIdentityAttributes] Object containing attributes of the client application identity or nil if the application identity is not available.
        def self.get_client_app_identity_attributes
          session = load_session();
          return session.get_client_app_identity_attributes();
        end

        # Return application identity attributes.
        # @private
        def self.get_app_identity_attrs(token)
          result = nil;

          claims = parse_claims(token);
          if ( claims != nil )
            result = UU::OS::Security::SessionAppIdentityAttributes.new();
            result.universe_id = claims[@@CLIENT_UID];
            result.app_code = claims[@@CLIENT_CODE];
          end

          return result;
        end
        private_class_method :get_app_identity_attrs;

        # @private
        def self.parse_claims(jwt_token)
          result = nil
          token = jwt_token
          if ( is_jwt(token) )
            token = token[@@JWT_TOKEN_PREFIX.length, token.length]
            parts = token.split(@@JWT_TOKEN_PARTS_SEPARATOR);
            if ( parts.length == 3 ) # three parts are expected, the second part contains claims
              b64value = add_base64_padding(parts[1]); # add missing padding characters to be divisible by 4
              json = Base64.decode64(b64value);
              result = JSON.parse(json);
            end
          end
          return result;
        end
        private_class_method :parse_claims;

        # @private
        def self.add_base64_padding(value)
          result = value
          # JRuby cannot decode value without the right padding (see https://github.com/jruby/jruby/issues/2815)
          if ( value && (value.length % 4 != 0) ) # each base64 encoded value length must be divisible by 4
            result = value << ("=" * (4 - (value.length % 4))); # append padding (concatenated '=' character) to conform to be divisible by 4
          end
          return result
        end
        private_class_method :add_base64_padding;

        # ===============================================================================
        # ==================== Session extension for commands v3 ========================
        # ===============================================================================

        # Creates new instance of session.
        # @param [String] access_token Access token.
        # @private
        def initialize(access_token) # :nodoc:
          @access_token = access_token
        end
        private_class_method :new

        # Returns current session (session object representing state after
        # most recent invocation of static +#login+ method).
        # @return [UU::OS::Security::Session] Session object.
        def self.current_session
          result = load_session();
          return result;
        end

        # Returns identity of the current application. Returns nil if the application identity is not available.
        # @return [UU::OS::Security::SessionAppIdentityAttributes] Object containing attributes of the server application identity or nil if the application identity is not available.
        def get_app_identity_attributes
          return self.class.send(:get_app_identity_attrs, @access_token);
        end

        # Returns identity of the client application. Returns nil if the application identity is not available.
        # @return [UU::OS::Security::SessionAppIdentityAttributes] Object containing attributes of the client application identity or nil if the application identity is not available.
        def get_client_app_identity_attributes
          result = nil;
          if ( @original_client_uid || @original_app_code )
            result = UU::OS::Security::SessionAppIdentityAttributes.new();
            result.universe_id = @original_client_uid;
            result.app_code = @original_app_code;
          end
          return result;
        end

        # Log user in. Unlike calling static +#login+, using this method
        # does not add new token into authentication stack and therefore
        # does not modify global context (returned session object must
        # be used for command invocation in order to invoke command as
        # newly authenticated user).
        #
        # @param [*String, Hash] credentials Valid parameters are:
        #   * Security realm, access code 1, access code 2
        #   * access code 1, access code 2 (uses default security realm)
        #   * Path to password file (absolute or relative to uu home)
        #   * HTTP authorization token
        #   * Hash with user (:user_credentials) and application (:app_credentials)
        #     credentials. Both values are optional and if specified, it can be
        #     formed from parameters valid for non-hash parameter variants.
        #     If no application credentials are given, the system will check whether
        #     exist the <%uu-home%>/app_identity file and loads application
        #     credentials from the file.
        #
        # @example
        #   credentials = Hash.new
        #   credentials[:user_credentials] = ['accessCode1', 'accessCode2']
        #   credentials[:app_credentials] = ['/etc/password_file']
        #   UU::OS::Security::Session.login(credentials)
        #
        # @return [UU::OS::Security::Session] New session object.
        def login(*credentials)
          session = self.class.send(:process_login, @access_token, *credentials);
          session.send(:previous_session, self);
          return session;
        end

        # Returns personal role URI of the authenticated user.
        #
        # @return [UU::OS::UESURI] URI of personal role
        def get_personal_role
          svc = UU::OS::REST::RemoteClient.new(Session)
          svc.auth_token = @access_token
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.get(:getPersonalRole)
            return UU::OS::UESURI.new(res)
          end
        end

        # Returns personal role URI of user who initiated session (created the first session object).
        #
        # @return [UU::OS::UESURI] URI of initiator personal role
        def get_initiator_personal_role
          svc = UU::OS::REST::RemoteClient.new(Session)
          svc.auth_token = @access_token
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.get(:getInitiatorPersonalRole)
            return UU::OS::UESURI.new(res)
          end
        end

        # Returns access role URI of authenticated user in given territory.
        # May return +nil+ in case user has no access role in territory.
        #
        # @param [String, UU::OS::UESURI] territory_uri URI of territory for which to get access role
        # @return [UU::OS::UESURI] URI of access role or +nil+
        def get_access_role(territory_uri)
          svc = UU::OS::REST::RemoteClient.new(Session)
          svc.auth_token = @access_token
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.get(:getAccessRole, territory_uri)
            if (res && res != @@NULL_URI_VALUE)
              return UU::OS::UESURI.new(res)
            else
              return nil
            end
          end
        end

        # Returns access role URI of user who initiated session (created the first session object)
        # in given territory. May return +nil+ in case user has no access role in territory.
        #
        # @param [String, UU::OS::UESURI] territory_uri URI of territory for which to get access role
        # @return [UU::OS::UESURI] URI of initiator access role or +nil+
        def get_initiator_access_role(territory_uri)
          svc = UU::OS::REST::RemoteClient.new(Session)
          svc.auth_token = @access_token
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.get(:getInitiatorAccessRole, territory_uri)
            if (res && res != @@NULL_URI_VALUE)
              return UU::OS::UESURI.new(res)
            else
              return nil
            end
          end
        end

        # Returns session access token.
        # @return [String] Access token.
        # @private
        def access_token # :nodoc:
          @access_token
        end

        # Override inspect to hide access token.
        # @private
        def inspect # :nodoc:
          "#<#{self.class.name}:0x#{(self.object_id << 1).to_s(16)}>"
        end

        private

        # Set incoming authentication token.
        #
        # @private
        def set_incoming_authn_token(token)
          if ( self.class.send(:is_jwt, token) )
            claims = self.class.send(:parse_claims, token);
            if ( claims != nil )
              @original_client_uid = claims[@@CLIENT_UID];
              @original_app_code = claims[@@CLIENT_CODE];
            end
          end
        end

        # Used for chaining information from previous session instance.
        #
        # @private
        def previous_session(previous_session)
          if ( previous_session != nil && get_client_app_identity_attributes() == nil )
            attrs = previous_session.get_client_app_identity_attributes();
            if ( attrs != nil )
              @original_client_uid = attrs.universe_id;
              @original_app_code = attrs.app_code;
            end
          end
        end

      end
    end
  end
end
